# encoding:utf8
import json
import os
import tkinter
import tkinter.font
import tkinter.ttk
import tkinter.messagebox
import tkinter.filedialog
from typing import List, Any, Union

from tooly.types import Literal

"""
此文件能实现通过简单的函数调用，实现基本的参数设置。
参数的类型支持：SingleChoiceParam（单选）、MultiChoiceParam（多选）、FileParam（文件）、TextParam（文本类型）以及ask_option(弹窗）
具体使用方法见代码中__main__
"""


class Param:

    def __init__(self, label: str, defaultVal: Any, validator, escapeCache):
        self.label = label  # 变量的名称
        self.defaultVal = defaultVal
        self.var = None
        self.validator = validator
        self.escapeCache = escapeCache

    def put_items(self, root, row):
        """
        :param root: 父容器（采用的是grid布局）
        :param row: 行号
        """
        raise NotImplementedError()

    def value(self):
        # 返回变量对应的value
        raise NotImplementedError()

    def valid(self) -> str:
        # get_args调用当前函数检查参数是否合法
        raise NotImplementedError()


class SingleChoiceParam(Param):
    def put_items(self, root, row):
        label = tkinter.Label(root, text=self.label + ":")
        label.grid(row=row, column=0, sticky='E')

        frame = tkinter.ttk.Frame(root)
        frame.grid(row=row, column=1, sticky='W')

        self.var = tkinter.IntVar()

        if self.defaultVal:
            self.var.set(self.options.index(self.defaultVal))
        else:
            self.var.set(0)
        for idx, option in enumerate(self.options):
            tmp = tkinter.Radiobutton(frame, text=option, variable=self.var, value=idx)
            tmp.grid(row=0, column=idx, sticky='W')

    def value(self):
        return self.options[self.var.get()]

    def valid(self) -> str:
        pass

    def __init__(self, label, options: List[str], defaultVal=None, escapeCache=False):
        if defaultVal is None:
            defaultVal = []
        super().__init__(label, defaultVal, None, escapeCache)
        self.options = options


class MultiChoiceParam(Param):
    def __init__(self, label: str, options: List[str], defaultVal: List[str] = None, notEmpty=False, escapeCache=False):
        if defaultVal is None:
            defaultVal = []
        super().__init__(label, defaultVal, None, escapeCache)
        self.options = options
        self.notEmpty = notEmpty

    def put_items(self, root, row):

        label = tkinter.Label(root, text=self.label + ":")
        label.grid(row=row, column=0, sticky='E')

        frame = tkinter.ttk.Frame(root)
        frame.grid(row=row, column=1, sticky='W')

        self.var = []

        for idx, option in enumerate(self.options):
            val = tkinter.IntVar()
            if option in self.defaultVal:
                val.set(1)
            self.var.append(val)
            w = tkinter.Checkbutton(frame, text=option, variable=val, onvalue=1, offvalue=0)
            w.grid(row=0, column=idx, sticky='W')

    def value(self):
        return [self.options[idx] for idx, i in enumerate(self.var) if i.get()]

    def valid(self) -> str:
        if self.notEmpty:
            if len(self.value()) == 0:
                return "不能为空"


class FileParam(Param):
    def valid(self) -> str:
        var = self.var.get()
        if self.required and not var:
            return "不能为空"

    def put_items(self, root, row):
        # 标签，父级元素为root
        label = tkinter.Label(root, text=self.label + ":")
        # 输入框绑定到的一个变量
        self.var = tkinter.StringVar(value=self.defaultVal)
        # 输入框
        text = tkinter.Entry(root, textvariable=self.var, state='disabled')
        dialog_type = self.dialog_type

        def fdialog():
            t = getattr(tkinter.filedialog, dialog_type)()
            self.var.set(t)

        btn = tkinter.ttk.Button(root, text="…", command=fdialog, width=2)

        # 标签通过grid方式布放，row为上层传递过来的，靠左
        label.grid(row=row, column=0, sticky="E", padx=5, pady=3)
        # 输入框靠右放
        text.grid(row=row, column=1, sticky="E")
        btn.grid(row=row, column=2, sticky='E', padx=3)

    def value(self):
        return self.var.get()

    def __init__(self, label, defaultVal='', required=True,
                 dialogType: Literal['askopenfilename', 'asksaveasfilename', 'askdirectory'] = 'askopenfilename',
                 escapeCache=False):
        super().__init__(label, str(defaultVal), None, escapeCache)
        self.dialog_type = dialogType
        self.required = required


class TextParam(Param):

    def __init__(self, label, defaultVal: Union[str, int] = "", validator=None, notBlank=False, isNumber=False,
                 escapeCache=False):
        super().__init__(label, str(defaultVal), validator, escapeCache)
        self.not_blank = notBlank
        self.require_number = isNumber

    def put_items(self, root, row):
        # 标签，父级元素为root
        label = tkinter.Label(root, text=self.label + ":")
        # 输入框绑定到的一个变量
        self.var = tkinter.StringVar(value=self.defaultVal)
        # 输入框
        text = tkinter.Entry(root, textvariable=self.var)

        # 标签通过grid方式布放，row为上层传递过来的，靠左
        label.grid(row=row, column=0, sticky="E", padx=5, pady=3)
        # 输入框靠右放
        text.grid(row=row, column=1, sticky="W")

    def value(self):
        if self.require_number:
            return int(self.var.get())
        return self.var.get()

    def valid(self):
        if self.validator:
            tmp = self.validator(self.var.get())
            if tmp:
                return tmp
        if self.not_blank and (len(self.var.get()) <= 0):
            return "不能为空"
        if self.require_number:
            try:
                int(self.var.get())
            except:
                return "需要为数字"


def __set_center(root):
    # 将弹框的位置设置到正中心
    root.update()
    curWidth = root.winfo_width()
    curHight = root.winfo_height()
    scn_w, scn_h = root.maxsize()
    # 计算中心坐标
    cen_x = (scn_w - curWidth) / 2
    cen_y = (scn_h - curHight) / 2
    # 设置窗口初始大小和位置
    size_xy = '%dx%d+%d+%d' % (curWidth, curHight, cen_x, cen_y)
    root.geometry(size_xy)


def get_args(args: List[Param], cache_file='args.json', 允许退出程序=True):
    """
    弹出框让用户设置参数，
    :return: 所有参数结果的list，如果用户点击关闭则自动退出程序
    """
    argNames = ''.join([item.label for item in args])
    cache_data = {}
    if cache_file and os.path.exists(cache_file):
        with open(cache_file, 'rb') as fd:
            cache_data = json.load(fd)
            data = cache_data.get(argNames, {})
            for arg in args:
                if arg.label in data and not arg.escapeCache:
                    arg.defaultVal = data[arg.label]

    root = tkinter.Tk()
    root.title("设置参数")
    root.resizable(0, 0)
    title = tkinter.Label(root, text="设置参数", font=tkinter.font.Font(family='microsoft yahei', size=14), pady=15)
    title.grid(row=0, column=0)
    frame = tkinter.ttk.Frame(root, padding=(30, 0, 30, 10))

    for (idx, param) in enumerate(args):
        param.put_items(frame, idx)

    _ret = {"ret": "exit"}

    def confirm():
        # 点击确定按钮后，检查所有的参数，并且收集所有的结果
        tmp = []
        for arg in args:
            msg = arg.valid()
            if msg:
                tkinter.messagebox.showwarning("参数不正确", arg.label + msg)
                return
            else:
                tmp.append(arg.value())
        _ret["ret"] = tmp
        root.destroy()

    btn = tkinter.Button(frame, text="确定", padx=20, command=confirm)
    btn.grid(row=len(args), column=1, sticky="W", pady=10)
    frame.grid(row=1, column=0)

    __set_center(root)
    if not 允许退出程序:
        root.protocol("WM_DELETE_WINDOW", lambda: False)
    root.mainloop()

    if _ret["ret"] == "exit":
        os._exit(0)

    tmp = {}
    for idx, p in enumerate(args):
        tmp[p.label] = _ret["ret"][idx]

    if cache_file:
        cache_data[argNames] = tmp
        jsonStr = json.dumps(cache_data, ensure_ascii=False, indent=4)
        with open(cache_file, 'w', encoding='utf8') as fd:
            fd.write(jsonStr)

    return tmp


def ask_option(title: str, options: List[str], 允许退出程序=False):
    root = tkinter.Tk()
    root.title(title)

    root.resizable(0, 0)
    _title = tkinter.Label(root, text=title, font=tkinter.font.Font(family='microsoft yahei', size=14), pady=30,
                           padx=30)
    _title.grid(row=0, column=0, sticky='w')

    frame = tkinter.ttk.Frame(root, padding=(15, 0, 15, 30))
    _ret = {}

    def handler(option):
        def ret():
            _ret["ret"] = option
            root.destroy()

        return ret

    for idx, option in enumerate(options):
        btn = tkinter.Button(frame, text=option, padx=15, command=handler(option))
        btn.grid(row=0, column=idx, padx=15)

    frame.grid(row=1, column=0)
    __set_center(root)
    if not 允许退出程序:
        root.protocol("WM_DELETE_WINDOW", lambda: False)
    root.mainloop()
    # 如果主循环关闭了，还没有设置选项，那就是因为点击了X导致的
    if len(_ret) == 0:
        os._exit(0)
    return _ret["ret"]


if __name__ == '__main__':
    # ret = get_args([TextParam("姓名说带飞时代发", notBlank=True, defaultVal="sssssssss"),
    #                 TextParam("年龄", isNumber=True, defaultVal=3),
    #                 MultiChoiceParam("duoxuan", ["苹果", "梨子"], defaultVal=["苹果"], notEmpty=False),
    #                 SingleChoiceParam("玩具", ["dao", "bi"], defaultVal="bi"),
    #                 FileParam("文件", required=False)], 允许退出程序=False)
    # print("ret = ", ret)
    #
    ret = get_args([TextParam("key", notBlank=True, defaultVal="sssssssss"),
                    TextParam("年龄", isNumber=True, defaultVal=3),
                    MultiChoiceParam("duoxuan", ["苹果", "梨子"], defaultVal=["苹果"], notEmpty=False),
                    SingleChoiceParam("玩具", ["dao", "bi"], defaultVal="bi"),
                    FileParam("文件", required=False)])

    print("ret2 = ", ret)

    ask_option("事实上事实上事实上", ["确 定"])



import argparse
import inspect
import json
import os
import subprocess
import pymysql
import sys
import zipfile
from typing import Union, Dict, Callable, List

from tooly.types import Encoding, Literal

defaultShellEncoding: Encoding = 'utf8'
if sys.platform.find('win'):
    defaultShellEncoding = 'gbk'

isDir = os.path.isdir


def getCallerFolder(depth=0):
    """
    :param depth: 调用者深度，默认为0层——调用者的文件夹位置  1层为调用者的调用者的位置
    """
    # 获取当前帧的上一层帧（即调用者的帧）
    caller_frame = inspect.stack()[1 + depth]
    # 获取该帧对应的文件路径
    caller_file_path = caller_frame.filename
    # 获取该文件的目录
    caller_folder = os.path.dirname(caller_file_path)
    return caller_folder


def run(cmd, encoding=defaultShellEncoding, failExit=True, withRetCode=False, log=print, showCmd=True, showOutput=True,
        cwd=None, noStdErr=False, returnJson=False):
    if not log:
        log = lambda *args: None
    if isinstance(cmd, list):
        if showCmd:
            log("==>>", ' \\\n\t'.join(cmd))
        cmd = ' '.join(cmd)
    else:
        if showCmd:
            log("==>>", cmd)
    errOut = None
    if noStdErr:
        errOut = subprocess.DEVNULL
    p = subprocess.Popen(cmd, shell=True, encoding=encoding, stdout=subprocess.PIPE, stderr=errOut, cwd=cwd)
    cmd = " ".join((i.strip("\\") for i in cmd.splitlines()))
    contents = []
    try:
        while p.poll() is None:
            line = p.stdout.readline()
            if showOutput:
                log(line, end="")
            contents.append(line)
    except KeyboardInterrupt:
        sys.exit()
    if failExit and p.returncode != 0:
        log(f'{cmd} 执行失败, code = {p.returncode}, stdout = \n {"".join(contents)}')
        sys.exit()
    retStr = "".join(contents).strip()
    if withRetCode:
        if returnJson:
            return p.returncode, json.loads(retStr)
        return p.returncode, retStr
    else:
        if returnJson:
            return json.loads(retStr)
        else:
            return retStr


def runS(cmd, encoding=defaultShellEncoding, failExit=True, withRetCode=False, log=print, showCmd=False,
         showOutput=False):
    """
    以silent的模式运行命令，仅返回结果
    """
    return run(cmd, encoding=encoding, failExit=failExit, withRetCode=withRetCode, log=log, showCmd=showCmd,
               showOutput=showOutput)


def runMysql(sql: str, host: str, user: str, password: str, port=3306, database=None, fetchAll=True, showSql=True):
    config = {
        "host": host,
        "user": user,
        "password": password,
        "port": port,
        "charset": "utf8mb4",
        "autocommit": True,
    }
    if database:
        config["database"] = database
    with pymysql.connect(**config) as connection:
        with connection.cursor() as cursor:
            # 执行SQL语句
            if showSql:
                print(sql)
            cursor.execute(sql)

            if not fetchAll:
                return
            # 获取查询结果的所有行
            results = cursor.fetchall()
            columns = [desc[0] for desc in cursor.description]
            # 返回所有的列名以及结果
            return columns, results


def runDocker(dockerName: str, image: str,
              removeAfterExit=True,
              daemon=True,
              dockerArgs: List[str] = None,
              removeOldDocker=True,
              network: str = None,
              ipAddr: str = None,
              portMappings: Dict[str, str] = None,
              envMappings: Dict[str, str] = None,
              curDirVolumeMappings: Dict[str, str] = None,
              volumeMappings: Dict[str, str] = None):
    if removeOldDocker:
        run(f"docker rm -f {dockerName}")

    if daemon:
        daemon = "-d"
    else:
        daemon = ""

    if not dockerArgs:
        dockerArgs = []

    removeAfterExitStr = ""
    if removeAfterExit:
        removeAfterExitStr = "--rm"

    portMappingsStr = []
    if portMappings:
        for k, v in portMappings.items():
            portMappingsStr.append(f"-p {k}:{v}")

    envMappingStr = []
    if envMappings:
        for k, v in envMappings.items():
            envMappingStr.append(f"-e {k}={v}")

    volumeMappingsStr = []
    if volumeMappings:
        for k, v in volumeMappings.items():
            volumeMappingsStr.append(f"-v {k}:{v}")

    if curDirVolumeMappings:
        for k, v in curDirVolumeMappings.items():
            k = os.path.normpath(os.path.join(getCallerFolder(1), k))
            volumeMappingsStr.append(f"-v {k}:{v}")
    networkStr = []
    if network:
        networkStr = ["--network " + network]

    if ipAddr:
        ipAddr = ['--ip=' + ipAddr]
    else:
        ipAddr = []

    cmd = [
        f"docker run --name {dockerName} {daemon} {removeAfterExitStr}",
        *networkStr,
        *ipAddr,
        *portMappingsStr,
        *envMappingStr,
        *volumeMappingsStr,
        f"{image}",
        *dockerArgs
    ]

    run(cmd)

    try:
        if daemon:
            res = json.loads(run("docker inspect " + dockerName, showOutput=False, showCmd=False))
            ip = res[0]["NetworkSettings"]["Networks"]["IPAddress"]
            print(f"{dockerName} IP: ", ip)
            return ip
    except:
        pass


class CustomHelpFormatter(argparse.HelpFormatter):
    """
    argparse的自定义help信息生成器，能在parser.add_argument函数没有help信息时打印default信息
    使用时，只需要在创建parser时传入即可，示例如下：
    >>> parser = argparse.ArgumentParser(formatter_class=CustomHelpFormatter)
    """

    def _format_args(self, action: argparse.Action, default_metavar: str) -> str:
        if action.default != argparse.SUPPRESS and action.default:
            default_val = "{" + f"{default_metavar}:{action.default}" + "}"
            return super()._format_args(action, default_val)
        else:
            return super()._format_args(action, default_metavar)


def writeText(fname: str, *contents: str, encoding: Encoding = "utf8"):
    mode = 'w'
    with open(fname, mode, encoding=encoding) as fd:
        for content in contents:
            fd.write(content)


def appendText(fname: str, *content: str, encoding: Encoding = 'utf8'):
    with open(fname, 'a+', encoding=encoding) as fd:
        for c in content:
            fd.write(c)


def readAsBytes(fname: str) -> bytes:
    with open(fname, 'rb') as fd:
        return fd.read()


def readAsText(fname: str, encoding: Encoding = "utf8") -> str:
    with open(fname, 'r', encoding=encoding) as fd:
        return fd.read()


def zipFiles(dst: str, *srcFiles: str):
    """
    :param dst: 压缩的目标文件
    :param srcFiles: 要压缩文件列表，支持文件通配符。需要传递相对路径否则会在压缩文件中创建全路径
    :return:
    """
    run(f"tar czf {dst} {' '.join(srcFiles)}")


def unzip(zipFname: str, dstDir: str):
    with zipfile.ZipFile(zipFname) as fd:
        fd.extractall(dstDir)


def p7zip(dst: str, *srcFiles: str):
    run(f"7z a {dst} {' '.join(srcFiles)}")


def p7unzip(zipFname: str, dstDir: str = ""):
    if dstDir:
        dstDir = "-o " + dstDir
    run(f"7z x {zipFname} {dstDir}")


def gitCommitId(requireClean: bool = True, shortId: bool = False, branch: str = "") -> Union[str, None]:
    """
    获取当前文件夹下的git commit id
    :param requireClean: 是否要求git status状态为clean
    :param shortId: 是否返回短ID（8字节）
    :param branch:若不为空，则要求当前的branch为指定branch
    :return:
    """
    if requireClean:
        ret = run("git status")
        if ret.find("nothing to commit, working tree clean") < 0:
            print("git代码未commit")
            return None

    if branch:
        if branch != gitBranch():
            print("git分支与预期不符")
            return None

    ret = run("git rev-parse HEAD")
    if shortId:
        return ret.strip()[:8]
    return ret.strip()


def gitBranch():
    ret = run("git branch")
    lines = ret.splitlines()
    for line in lines:
        line = line.strip()
        if line.startswith("*"):
            return line.split(" ")[1]


def genDockerRunCmd(dockerName: str,
                    dockerImage: str,
                    command: str = "",
                    customParams: str = "",
                    daemomMode=False,
                    envs: Dict[str, str] = {},
                    fileMapping: Dict[str, str] = {},
                    portMapping: Dict[int, int] = {},
                    workingDirInDocker: str = None,
                    multiLine: bool = False,
                    removeOnExit: bool = True):
    """
    生成docker运行的命令
    :param dockerName: docker名称
    :param dockerImage: docker image
    :param command: 可选，docker的启动命令
    :param customParams: 额外的命令行参数可以在此指定
    :param daemomMode: 是否在后台运行
    :param envs: 环境变量列表
    :param fileMapping: 文件映射列表 host -> docker
    :param portMapping: 端口映射列表 host -> docker
    :param workingDirInDocker: docker内pwd
    :param removeOnExit: docker结束后是否删除docker
    :param multiLine 命令是否分割成多行
    :return:
    """
    if not dockerImage:
        raise Exception("docker image不能为空")
    if not dockerName:
        raise Exception("docker name不能为空")

    _removeOnExit = " --rm" if removeOnExit else ""
    _daemonMode = " -d" if daemomMode else ""

    lineStart = "\t" if multiLine else ""
    lineEnd = " \\\n" if multiLine else ""

    mappings = []
    mappings += [f'{lineStart}-e {k}={v}{lineEnd}' for (k, v) in envs.items()]
    mappings += [f'{lineStart}-v {k}:{v}{lineEnd}' for (k, v) in fileMapping.items()]
    mappings += [f'{lineStart}-p {k}:{v}{lineEnd}' for (k, v) in portMapping.items()]
    _workdingDir = f" -w {workingDirInDocker}" if workingDirInDocker else ""

    _mappings = " ".join(mappings)

    return f"docker run{_removeOnExit}{_daemonMode}{customParams} --name {dockerName}{lineEnd}{_mappings}{lineStart}{_workdingDir}{dockerImage}{command}"


def runInTmux(cmd: str, sessionName: str = None):
    """
    新建一个tmux，并在其中运行命令（python脚本退出后，tmux中的命令仍然在运行
    """
    try:
        import libtmux
    except:
        raise Exception("libtmux未安装，请使用如下命令安装：pip install libtmux")
    server = libtmux.Server()
    if sessionName:
        sess = server.find_where({"session_name": sessionName})
        if sess:
            sess.kill_session()
    sess = server.new_session(session_name=sessionName)
    sess.attached_pane.send_keys(cmd.strip() + "\n")


def sshExecWithParamiko(cmd: str):
    try:
        import paramiko
    except:
        raise Exception("libtmux未安装，请使用如下命令安装：pip install paramiko")


def withParamiko(host: str, userName: str, fun: Callable,
                 port: int = 23,
                 isScp: bool = False,
                 passwd: str = None,
                 rsaFile: str = None,
                 rsaPasswd: str = None):
    try:
        import paramiko
        from scp import SCPClient
    except:
        raise Exception("libtmux未安装，请使用如下命令安装：pip install paramiko")

    ssh = paramiko.SSHClient()
    try:
        ssh.load_system_host_keys()
        pkey = None
        if rsaFile:
            pkey = paramiko.RSAKey.from_private_key_file(filename=rsaFile, password=rsaPasswd)
        ssh.connect(hostname=host, port=port, username=userName, pkey=pkey, password=passwd)

        if isScp:
            with SCPClient(ssh.get_transport()) as scp:
                fun(scp)
        else:
            fun(ssh)

    finally:
        ssh.close()


def scpWithParamiko(srcPath: str,
                    dstPath: str,
                    host: str,
                    userName: str,
                    port: int = 23,
                    passwd: str = None,
                    rsaFile: str = None,
                    rsaPasswd: str = None,
                    recursive: bool = True,
                    method: Literal["put", "get"] = "put"):
    def f(scp):
        if "put" == method:
            scp.put(srcPath, dstPath, recursive=recursive)
        else:
            scp.get(srcPath, dstPath, recursive=recursive)

    withParamiko(host, userName, f, isScp=True, port=port, passwd=passwd, rsaFile=rsaFile, rsaPasswd=rsaPasswd)


if __name__ == '__main__':
    # print(ls("*.py"))
    # touch("a/b/c.txt")
    # move("a/*.txt","a/b")
    # rm("a/b/*.txt")
    # print(ls("a/b/"))
    # rm("a")
    # print(ls("a/b/"))
    # touch("a/b/c.txt")
    # touch("a/b.txt")
    # mkdir("a/d")
    # zip("a/", "f.zip")
    # zip("f.zip", "a", excludeFiles=["a/d"])
    # touch("a/b/b1.txt")
    # mkdir("a/b/c/")
    # touch("a/d.txt")
    # mkdir("a/d/")
    # p7zip("out.7z", "a")
    # unzip("out.zip", "extract")
    # p7unzip("out.7z", "extract")
    # print(gitCommitId(requireClean=False))
    # print(genDockerRunCmd("ddd", "nginx", fileMapping={
    #     "/hostDir": "/dockerDir",
    #     "/hostDir2": "/dockerDir"
    # }, portMapping={
    #     22: 33,
    #     44: 55,
    #     23: 4403
    # }, envs={
    #     "NAME": "liyuan"
    # }, multiLine=True))
    run("ls")



import math
from threading import Thread
from typing import List, Any, Callable


class Result:
    def __init__(self, result: Any, exception: Exception = None):
        self.result = result
        self.exception = exception

    def __repr__(self):
        return f"[result = {self.result}, exception = {self.exception}]"


def runInMultiThread(threadCnt: int, func: Callable, args: List[List[Any]]) -> List[Result]:
    """
    启动threadCnt个线程，循环运行func函数
    :param threadCnt: 线程数量
    :param func: 线程中要运行的函数
    :param args: List<List>
    :return: List<Result>
    """
    threadResultList = [None] * threadCnt

    def f(i, arg):
        resList = []
        for a in arg:
            try:
                res = func(*a)
                resList.append(Result(res))
            except Exception as e:
                resList.append(Result(None, e))
        threadResultList[i] = resList

    thList = []
    argSize = math.ceil(len(args) / threadCnt)
    for i in range(threadCnt):
        th = Thread(target=f, args=(i, args[i * argSize:(i + 1) * argSize]))
        th.start()
        thList.append(th)

    for th in thList:
        th.join()
    result = []
    for res in threadResultList:
        result += res
    return result


if __name__ == '__main__':
    def f(i):
        return f"i = {i}, 1/i = {1 / i}"


    thArgs = [[1], [2], [3], [4], [0]]

    for i in runInMultiThread(3, f, thArgs):
        print(i)



import csv
from typing import List


def writeCsv(fname: str, datas: List[List[str]], encoding: str = 'utf8'):
    with open(fname, 'w', encoding=encoding) as fd:
        writer = csv.writer(fd)
        writer.writerows(datas)
